/*	This is pstree V1.4 written by Fred Hucht (c) 1993	*
 *	EMail: fred@hal6000.Uni-Duisburg.DE			*
 *	Feel free to copy and redistribute in terms of the	*
 * 	GNU public license. 					*/

static char *WhatString[2]= {
  "@(#)pstree V1.4 by Fred Hucht (C) 1993",
  "@(#)EMail:fred@thp.Uni-Duisburg.DE"
  };

#define MAXLINE 256

#if defined(_AIX) || defined(___AIX)	/* AIX 3.x */
#  define HAS_TERMDEF
#  define PSCMD "ps -ekf"
#  define PSFORMAT "%s %d %d %*20c %*s %[^\n]"
#elif defined(__linux)			/* Linux 0.9x */
#  define UID2USER
#  define PSCMD "ps laxw"
#  define PSFORMAT "%*d %d %d %d %*35c %*s %[^\n]"
#elif defined(_BSD)			/* Untested */
#  define UID2USER
#  define PSCMD "ps laxw"
#  define PSFORMAT "%*d %*c %d %d %d %*d %*d %*d %*x %*d %d %*15c %*s %[^\n]"
#elif defined(__convex)			/* Convex */
#  define UID2USER
#  define PSCMD "ps laxw"
#  define PSFORMAT "%*s %d %d %d %*d %*g %*d %*d %*21c %*s %[^\n]"
#else					/* HP-UX, A/UX etc. */
#  define PSCMD "ps -ef"
#  define PSFORMAT "%s %d %d %*20c %*s %[^\n]"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>		/* For str...() */
#include <unistd.h>		/* For getopt */

#ifdef UID2USER
#include <pwd.h>
struct passwd *pw;
#endif

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

struct TreeChars {
  char *s1, 		/* String to intend */
       *s2, 		/* String between header and pid */
       *p, 		/* dito, when parent of printed childs */
        c, 		/* Last char of s1 for line with child */
        e;		/* Last char of s1 for last child */
};

static struct TreeChars 
  Ascii = { " |",    "---",          "-+-",          '|',    '\\' },
  Pc850 = { " \263", "\304\304\304", "\304\302\304", '\303', '\300' },
  *c;

int MyPid, NProc, Ns1, Maxlevel = 0, Alignlen, Columns;
short align = FALSE, showall = TRUE, soption = FALSE;
char *name = "", *str = NULL;
long ipid = -1;

#ifdef DEBUG
int debug = FALSE;
#endif


struct { 
  long uid, pid, ppid;
  char name[9], cmd[MAXLINE];
  int  parent, child, pchilds, level;
} *p;

int getprocs() {
  FILE *tn;
  int len, i = 0;  extern int errno; /* For popen() */
#ifdef UID2USER
  int xx; /* For testing... */
#endif
  char line[MAXLINE], command[] = PSCMD;

  if(NULL == (tn = popen(command,"r"))) {
    fprintf(stderr, "Problems with pipe, errno = %d\n",errno);
    exit (1);
  }
#ifdef DEBUG
  if(debug) printf("popen:errno = %d\n",errno);
#endif
  
  fgets(line, MAXLINE, tn); /* Throw away header line */
#ifdef DEBUG
  if(debug) fputs(line, stdout);
#endif

  if ( NULL == (p = malloc( sizeof(*p) ))) {
    fprintf(stderr, "Problems with malloc.\n");
    exit (1);
  }
  
  while(NULL != fgets(line, MAXLINE, tn) && 10 < (len = strlen(line))) {
#ifdef DEBUG
    if(debug) {printf("len=%3d ",len); fputs(line, stdout);}
#endif
    
    if ( NULL == (p = realloc(p, (i+1) * sizeof(*p) ))) {
      fprintf(stderr, "Problems with realloc.\n");
      exit (1);
    }

#ifdef UID2USER 	/* read BSD-style ps output */
    sscanf(line, PSFORMAT, &p[i].uid, &p[i].pid, &p[i].ppid, p[i].cmd);
    pw = getpwuid(p[i].uid); 
    strcpy(p[i].name, pw->pw_name); 
#else			/* read SYSV-style ps output */
    sscanf(line, PSFORMAT, p[i].name, &p[i].pid, &p[i].ppid, p[i].cmd);
#endif

#ifdef DEBUG
    if(debug) printf("uid=%5d, name=%8s, pid=%5d, ppid=%5d, cmd='%s'\n", 
		     p[i].uid, p[i].name, p[i].pid, p[i].ppid, p[i].cmd);
#endif
    p[i].parent = p[i].child = FALSE;
    p[i].pchilds = 0;
    i++;
  }
  pclose(tn);
  return i;
}

#define IS_CHILD(i) (p[i].ppid == pid && p[i].pid != pid)

int pstree1(long pid, int level) {
  int i = 0, j;
  while(i <= NProc && p[i].pid != pid ) i++; /* Search process */
  if (i > NProc) return 1; /* Return if pid does not exist */
  
  if(0 == strcmp(p[i].name, name) 	/* for -u */
     || p[i].pid == ipid		/* for -p */
     || (soption
	 && NULL != strstr(p[i].cmd, str)
	 && p[i].pid != MyPid))		/* for -s */
    p[i].parent = p[i].child = TRUE;

  p[i].level = level;
  
  if(p[i].pid == MyPid) return 0; /* Ignore my children (sh, ps)... */

  for(j=0; j<NProc; j++) if (IS_CHILD(j)) { 
    p[j].child   = p[i].child;  /* set if parent user matches */
    if (pstree1(p[j].pid, level + 1))  /* call me */
      fprintf(stderr, "This should not happen. pid=%d\n",p[j].pid);
    p[i].parent |= p[j].parent; /* set if child user matches */
  }
  return 0;
}

#define IS_PRINTED(i) (showall || p[i].child || p[i].parent)

void pstree2(long pid, char *head) {
  int i = 0, j, headL, nheadL;
  char nhead[MAXLINE], out[MAXLINE], *Format;
  
  while(i <= NProc && p[i].pid != pid ) i++; /* Search process */
  if (i > NProc) { 
    fprintf(stderr, "This should not happen (2). pid=%d\n", pid);
    exit(1);
  }

  if (!IS_PRINTED(i)) return; /* No need to process */
  
  for(j=0; j<NProc; j++) if (IS_CHILD(j) && IS_PRINTED(j)) 
    p[i].pchilds++; /* Count printed childs of i */
  
  if(p[i].pid == MyPid) p[i].pchilds = 0; /* Ignore my childs */
  
  headL = Ns1 * p[i].level - 1; /* = strlen(head)-1; */

  if(align) {
    for(j = 0; j < Alignlen - headL; j++) nhead[j] = c->s2[0];
    nhead[j] = '\0';
    Format = "%s%s%s %05d %-8s %s";
  }
  else {
    nhead[0] = '\0';
    Format = "%s%s%s %05d %s %s";
  }
  
  sprintf(out, Format, head, p[i].pchilds ? c->p : c->s2,
	  nhead, pid, p[i].name,p[i].cmd);
  out[headL] = (out[headL] == c->e) ? c->e : c->c;
  out[Columns] = '\0';
  puts(out);
  
  if(p[i].pid == MyPid) return; /* Don't show my childs... */
  
  if(head[headL] == c->e) head[headL] = ' ';

  sprintf(nhead,"%s%s", head, c->s1);
  nheadL = headL + Ns1; /* = strlen(nhead)-1; */

  for(j=0; j<NProc; j++) if (IS_CHILD(j)) { 
    if(IS_PRINTED(j) && 1 == p[i].pchilds--) 
      nhead[nheadL] = c->e; /* last child gets c->e */
    pstree2(p[j].pid, nhead);
  }
}

void Pstree(long pid) {
  int i, r;
  
  r = pstree1(pid, 0); /* Pass 1 */
  
  if (r) { 
    fprintf(stderr, "Process %d does not exist!\n",pid);
    return;
  }

  if(align) { /* Shall output be aligned? */
    for(i = 0; i < NProc; i++) 
      if (IS_PRINTED(i) && Maxlevel < p[i].level) 
	Maxlevel = p[i].level;
    Alignlen = Maxlevel * Ns1 - 1;
#ifdef DEBUG
    if(debug) printf("Maxlevel = %d, Alignlen = %d\n",Maxlevel, Alignlen);
#endif
  }
  
  pstree2(pid, ""); /* Pass 2 */
}

void Usage(char *progname) {
  char *p;
  p = (NULL == (p = strrchr(progname,'/'))) ? progname : p+1;
  fprintf(stderr,
	  "%s, %s\n\n"
	  "Usage: %s [-a] [-g] [-u user] [-s string] [-p pid] [-w] [pid ...]\n"
	  "   -a        align output\n"
	  "   -g        use IBM-850 graphics chars for tree\n"
	  "   -u user   show only parts containing processes of <user>\n"
          "   -s string show only parts containing process with <string> in commandline\n"
          "   -p pid    show only parts containing process <pid>\n"
	  "   -w        wide output, not truncated to window width\n"
	  "   pid ...   process ids to start from, default 1\n"
	  , WhatString[0]+4, WhatString[1]+4, p);
  exit(1);
}

int main(int argc, char **argv) {
  extern int optind;
  extern char *optarg;
  int ch;
  long pid;
  int graph = FALSE, wide = FALSE;

  while ((ch = getopt(argc, argv, "adghp:s:u:w?")) != EOF) 
    switch (ch) {
    case 'a': align   = TRUE; break;;
#ifdef DEBUG
    case 'd': debug   = TRUE; break;;
#endif
    case 'g': graph   = TRUE; break;;
    case 'p': showall = FALSE;
              ipid    = atoi(optarg); break;;
    case 's': showall = FALSE;
              soption = TRUE;
              str     = optarg; break;;
    case 'u': showall = FALSE;
              name    = optarg; break;;
    case 'w': wide    = TRUE; break;;
    case 'h':
    case '?':
    default : Usage(argv[0]); break;;
  }

  NProc = getprocs();
  MyPid = getpid();
  if(wide)
    Columns = MAXLINE-1;
  else {
#ifdef HAS_TERMDEF
    Columns = atoi((char*)termdef(fileno(stdout),'c'));
#else
    Columns = 80;
#endif
  }
  c = graph ? &Pc850 : &Ascii;

  Ns1 = strlen(c->s1);
  
#ifdef DEBUG
  if(debug) printf("Columns = %d\n", Columns);
#endif
  if(Columns == 0 || Columns >= MAXLINE) Columns = MAXLINE-1;
  if(argc == optind) /* No pids */
    Pstree(1);
  else while (optind < argc) {
    pid = (long)atoi(argv[optind]);
    Pstree(pid);
    optind++;
  }
  free(p);
  return 0;
}
